module PrioQueue(ListPrioQueue, empty, insert, getMin, deleteMin) where

import Control.Exception.Base (assert)

newtype ListPrioQueue a = PQ [a]

empty :: ListPrioQueue a
empty = PQ []

insert :: Ord a => a -> ListPrioQueue a -> ListPrioQueue a  -- O(n)
insert x pq@(PQ ys) =
 assert (invariant pq) $
 PQ $ ins ys
 where
  ins [] = [x]
  ins (y : ys) | x <= y = x : y : ys
  ins (y : ys) = y : ins ys

getMin :: ListPrioQueue a -> a  -- O(1)
getMin (PQ xs) = head xs

deleteMin :: Ord a => ListPrioQueue a -> (a, ListPrioQueue a)  -- O(1)
deleteMin pq@(PQ xs) =
 assert (invariant pq) $
 (head xs, PQ (tail xs))

invariant :: Ord a => ListPrioQueue a -> Bool
invariant (PQ xs) = ordered xs
 where
  ordered [] = True
  ordered [_] = True
  ordered (x : y : xs) =
   case compare x y of
    GT -> False
    _ -> ordered (y : xs)


